#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include "tron.h"

template <class T> static inline void swap(T* a, int x, int y) { T t=a[x]; a[x]=a[y]; a[y]=t; }

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#ifdef __cplusplus
extern "C" {
#endif

	extern double dnrm2_(int *, double *, int *);
	extern double ddot_(int *, double *, int *, double *, int *);
	extern int daxpy_(int *, double *, double *, int *, double *, int *);
	extern int dscal_(int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}

#define MEPS 1e-16L
#define PIVOTING
#define DEBUG_LDL

// solve Ax=b , in which A[i*n+j] = Aij and we only solve submatrix A[m:m]
double solve_system(const double *_A, const double *_b, int n, int m, double *x)
{
	long double *A = new long double[n * n];
	for(int i=0; i<n*n; i++)
		A[i] = _A[i];
	long double *b = new long double[n];
	for(int i=0; i<n; i++)
		b[i] = _b[i];
	for(int i=0; i<m; i++)
		x[i] = 0;

#ifdef PIVOTING
	int *p = new int[m];
	for(int i=0; i<m; i++)
		p[i] = i;
#endif

	// decomp
	// A -> D0
	//		L00 D1
	//		L10 L11 D2
	for(int j=0; j<m; j++){
#ifdef PIVOTING
		int pj = j;
		long double pj_val = fabsl(A[j*n+j]);
		// find piv
		for(int i=j+1; i<m; i++){
			long double pi_val = fabsl(A[i*n+i]);
			if(pj_val < pi_val){
				pj_val = pi_val;
				pj = i;
			}
		}
		// swap rows and columns of j and pj
		for(int i=0; i<m; i++)
			swap(A, j*n+i, pj*n+i);
		for(int i=0; i<m; i++)
			swap(A, i*n+j, i*n+pj);
		swap(b, j, pj);
		swap(p, j, pj);
#endif

		// loop invariant: Aik = Lik, Akk = Dk with k<j
		long double Dj = A[j*n+j];
		for(int k=0; k<j; k++){
			Dj -= A[j*n+k]*A[j*n+k] * A[k*n+k];
		}
		A[j*n+j] = Dj;
		//assert(fabs(Dj) >= MEPS);
		if(fabsl(Dj) < MEPS){
			for(int i=j+1; i<m; i++)
				A[i*n+j] = 0;
			continue; // xj should be 0
		}

		for(int i=j+1; i<m; i++){
			long double Lij = A[i*n+j];
			for(int k=0; k<j; k++){
				Lij -= A[i*n+k] * A[j*n+k] * A[k*n+k];
			}
			Lij /= Dj;
			A[i*n+j] = Lij;
		}
	}
#if 0
	printf("L\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<=i; j++)
			printf("%+.3Le ", A[i*n+j]);
		printf("\n");
	}
#endif
	// invariant: A[i*n+j] = Lij, forall i>j
	//			  A[i*n+i] = Dii, forall i

	// LD y = b
	// di yi =	bi - \sum_{j<i} yj Lij
	// place y in b
	for(int i=0; i<m; i++){
		// loop invariant: bk = yk forall k<i
		long double yi = b[i];
		for(int k=0; k<i; k++)
			yi -= A[i*n+k] * b[k];
		b[i] = yi;
	}
	for(int i=0; i<m; i++){
		long double Di = A[i*n+i];
		if(fabsl(Di) < MEPS)
			b[i] = 0;
		else
			b[i] /= Di;
	}
#if 0
	printf("y\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", b[i]);
	}
#endif
	// L'x = y
	// xi = yi - \sum{j>i} xj Lij
	for(int i=m-1; i>=0; i--){
		// loop invariant: xk ready forall k>i
		long double xi = b[i];
#ifdef PIVOTING
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[p[k]];
		x[p[i]] = (double)xi;
#else
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[k];
		x[i] = (double)xi;
#endif
	}
#if 0
	printf("\n");
	for(int i=0; i<n; i++)
		printf("d %.3e\n", x[i]);

	printf("A\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<n; j++)
			printf("%+.3e ", Acopy[i*n+j]);
		printf("\n");
	}
	printf("b\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", bcopy[i]);
	}
#endif

#if 1
	double err = 0;
	for(int i=0; i<m; i++){
		double Ax_i = 0;
		for(int j=0; j<m; j++)
			Ax_i += _A[i*n+j] * x[j];
		err += fabs(Ax_i-_b[i]);
	}
	double max_pv = 0;
	for(int i=0; i<m; i++)
		if(max_pv < _A[i*n+i])
			max_pv = (double)_A[i*n+i];
	if(max_pv != 0)
		err /= max_pv;
	//printf("err=%e\n", err);
	//assert(err < 1e-5);
#endif
	delete[] A;
	delete[] b;
#ifdef PIVOTING
	delete[] p;
#endif
	return err;
}

void TRON::info(const char *fmt,...)
{
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*tron_print_string)(buf);
}

TRON::TRON(
		const function *fun_obj,
		double eps, double eps_cg,
		int max_iter,
		int num_directions,
		int subsample_ratio, int max_cg_iter,
		double alpha_pcg,
		int seed)
{
	this->fun_obj=const_cast<function *>(fun_obj);
	this->eps=eps;
	this->eps_cg=eps_cg;
	this->max_iter=max_iter;
	this->num_directions = num_directions;
	this->subsample_ratio = subsample_ratio;
	this->max_cg_iter = max_cg_iter;
	this->alpha_pcg = alpha_pcg;
	tron_print_string = default_print;
	srand(seed);
}

TRON::~TRON()
{
}

void TRON::tron(double *w)
{
	const clock_t begin_time = clock();
	int l = fun_obj->get_nr_instance();
	int n = fun_obj->get_nr_variable();
	int i, cg_iter;
	double step_size;
	double f, fnew, actred;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];

	double *Hs = new double[n];
	double *Xs = new double[l];

	long long data_access = 0;
	int *sample_size = new int[subsample_ratio];

	double *D = new double[n];

	int limit = num_directions;
	if (limit == -1) {
		limit = (int)sqrt((double)fun_obj->get_nr_element()/l);
		if (limit%2 != 1) limit++;
	}

	double *P = new double[n*limit];
	double *XP = new double[l*limit];
	double *Pg = new double[limit];
	double *A = new double[limit*limit];
	double *inner = new double[limit*limit];
	double *b = new double[limit];
	double *t = new double[limit];

	info("l %d max_cg_iter %d num_directions %d\n", l, max_cg_iter, limit);
	for (i=0; i<subsample_ratio; i++) {
		sample_size[i] = l/subsample_ratio;
		if (i < l%subsample_ratio)
			sample_size[i]++;
	}

	// calculate gradient norm at w=0 for stopping condition.
	double *w0 = new double[n];
	for (i=0; i<n; i++)
		w0[i] = 0;
	fun_obj->fun(w0, data_access);
	fun_obj->grad(w0, g, data_access);
	double gnorm0 = dnrm2_(&n, g, &inc);
	delete [] w0;

	f = fun_obj->fun(w, data_access);
	fun_obj->grad(w, g, data_access);
	double gnorm = dnrm2_(&n, g, &inc);

	if (gnorm <= eps*gnorm0)
		search = 0;

	double gTs, wTs, sTs;
	int idx = 0, num = 0;

	while (iter <= max_iter && search)
	{
		clock_t clk = clock();
		int current_sample_size = sample_size[(iter-1)%subsample_ratio];
		fun_obj->fix_a_sample(current_sample_size);

		cg_iter = trpcg(g, D, s, w, r, f, data_access);

		double c;
		c  = 1/dnrm2_(&n, s, &inc);
		dscal_(&n, &c, s, &inc);

		fun_obj->Xv(s, Xs, data_access);

		memcpy(P+n*idx, s, sizeof(double)*n);
		memcpy(XP+l*idx, Xs, sizeof(double)*l);
		num = min(num+1, limit);

		for (i=0; i<num; i++) {
			int p = (idx+limit-i)%limit;
			b[i] = -ddot_(&n, g, &inc, P+n*p, &inc);
			inner[p*limit+idx] = inner[idx*limit+p] = ddot_(&n, P+n*p, &inc, P+n*idx, &inc);
		}

		fun_obj->get_matrix(XP, inner, A, limit, num, idx);
		solve_system(A, b, num, num, t);

		memset(s, 0, sizeof(double)*n);

		for (i = 0; i < num; i++) {
			int p = (idx+limit-i)%limit;
			daxpy_(&n, &t[i], P+n*p, &inc, s, &inc);
		}

		memset(Xs, 0, sizeof(double)*l);

		for (i = 0; i < num; i++) {
			int p = (idx+limit-i)%limit;
			daxpy_(&l, &t[i], XP+l*p, &inc, Xs, &inc);
		}

		// no need to increase num because of linear dependency when num < limit
		idx = (idx+1)%limit;

		// use the update direction
		memcpy(P+n*idx, s, sizeof(double)*n);
		memcpy(XP+l*idx, Xs, sizeof(double)*l);
		c  = 1/dnrm2_(&n, P+n*idx, &inc);
		dscal_(&n, &c, P+n*idx, &inc);
		dscal_(&l, &c, XP+l*idx, &inc);
		for (i=0; i<num; i++) {
			int p = (idx+limit-i)%limit;
			inner[p*limit+idx] = inner[idx*limit+p] = ddot_(&n, P+n*p, &inc, P+n*idx, &inc);
		}

		wTs = ddot_(&n, w, &inc, s, &inc);
		gTs = ddot_(&n, g, &inc, s, &inc);
		sTs = ddot_(&n, s, &inc, s, &inc);

		step_size = fun_obj->line_search(s, w, g, Xs, sTs, wTs, gTs, fnew, data_access);
		double prered = -0.5*gTs;
		actred = f - fnew;

		if (step_size == 0)
		{
			info("WARNING: line search fails\n");
			break;
		}

		fun_obj->step(sTs, wTs, step_size, fnew);
		daxpy_(&n, &step_size, s, &inc, w, &inc);

		double snorm = dnrm2_(&n, s, &inc);
		double wnorm = dnrm2_(&n, w, &inc);
		double time = double(clock() - begin_time) / CLOCKS_PER_SEC;

		idx = (idx+1)%limit;

		info("iter %2d f %.16f |g| %g CG %3d |s| %g step_size %g |w| %g wTs %g pre %.16f act %.16f num_data %lld approx_pass %lld num_sample %d time %.f\n",
			iter, f, gnorm, cg_iter, snorm,
			step_size, wnorm, wTs, prered, actred,
			data_access, data_access/l, current_sample_size, time);

		f = fnew;
		iter++;

		fun_obj->grad(w, g, data_access);

		gnorm = dnrm2_(&n, g, &inc);
		if (gnorm <= eps*gnorm0) {
			double f_check = fun_obj->fun(w, data_access);
			info("FINISH %.16f %.16f\n", f, f_check);
			break;
		}
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f))
		{
			info("WARNING: actred too small\n");
			break;
		}
	}

	delete[] g;
	delete[] r;
	delete[] s;
	delete[] sample_size;
	delete[] D;
	delete[] Hs;
	delete[] Xs;

	delete[] P;
	delete[] XP;
	delete[] Pg;
	delete[] A;
	delete[] inner;
	delete[] b;
	delete[] t;
}

int TRON::trpcg(double *g, double *M, double *s, double *w, double *r, double f, long long &data_access)
{
	int i, inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;

	int *D = new int[n];
	int *idx = new int[n];
	int *rev = new int[n];
	double *d = new double[n];
	double *Hd = new double[n];
	double zTr, znewTrnew, alpha, beta;

	fun_obj->get_use(D, data_access);

	int cnt = 0;
	double pTp = 0;
	double c_s = 0;
	double c_r = -1;
	double c_d = c_r;
	for (i=0; i<n; i++) {
		if (D[i] == 0) {
			idx[i] = -1;
			pTp += g[i]*g[i];
		} else {
			idx[i] = cnt;
			rev[cnt] = i;
			s[cnt] = 0;
			r[cnt] = -g[i];
			d[cnt] = r[cnt];
			cnt++;
		}
	}

	fun_obj->setup_use(idx, data_access);

	zTr = ddot_(&cnt, r, &inc, r, &inc)+pTp*c_r*c_r;
	int cg_iter = 0;

	while (cg_iter < this->max_cg_iter)
	{
		cg_iter++;

		memset(Hd, 0, sizeof(double)*cnt);
		fun_obj->sample_Hv(d, Hd, data_access);
		daxpy_(&cnt, &one, d, &inc, Hd, &inc);

		alpha = zTr/(ddot_(&cnt, d, &inc, Hd, &inc)+pTp*c_d*c_d);
		daxpy_(&cnt, &alpha, d, &inc, s, &inc);
		c_s += alpha*c_d;

		alpha = -alpha;
		daxpy_(&cnt, &alpha, Hd, &inc, r, &inc);
		c_r += alpha*c_d;

		znewTrnew = ddot_(&cnt, r, &inc, r, &inc)+pTp*c_r*c_r;
		beta = znewTrnew/zTr;
		dscal_(&cnt, &beta, d, &inc);
		c_d *= beta;
		daxpy_(&cnt, &one, r, &inc, d, &inc);
		c_d += c_r;
		zTr = znewTrnew;
	}

	for (i=0; i<cnt; i++)
		d[i] = s[i];

	for (i=0; i<n; i++) {
		if (idx[i] == -1) {
			s[i] = g[i]*c_s;
		} else {
			s[i] = d[idx[i]];
		}
	}

	delete[] D;
	delete[] idx;
	delete[] rev;
	delete[] d;
	delete[] Hd;

	return(cg_iter);
}

double TRON::norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax)
			dmax = fabs(x[i]);
	return(dmax);
}

void TRON::set_print_string(void (*print_string) (const char *buf))
{
	tron_print_string = print_string;
}
